/*
Copyright (C) BABEC. All rights reserved.
Copyright (C) THL A29 Limited, a Tencent company. All rights reserved.

SPDX-License-Identifier: Apache-2.0
*/

package vm

import (
	"errors"
	"fmt"
	"sort"
	"strconv"

	acPb "chainmaker.org/chainmaker/pb-go/v2/accesscontrol"
	"chainmaker.org/chainmaker/pb-go/v2/common"
	"chainmaker.org/chainmaker/protocol/v2"
)

const (
	constructKeySeparator = "#"
)

func NewTxSimContext(vmManager protocol.VmManager, snapshot protocol.Snapshot, tx *common.Transaction,
	blockVersion uint32) protocol.TxSimContext {
	return &txSimContextImpl{
		txExecSeq:        snapshot.GetSnapshotSize(),
		tx:               tx,
		txReadKeyMap:     make(map[string]*common.TxRead, 8),
		txWriteKeyMap:    make(map[string]*common.TxWrite, 8),
		sqlRowCache:      make(map[int32]protocol.SqlRows),
		kvRowCache:       make(map[int32]protocol.StateIterator),
		txWriteKeySql:    make([]*common.TxWrite, 0),
		txWriteKeyDdlSql: make([]*common.TxWrite, 0),
		snapshot:         snapshot,
		vmManager:        vmManager,
		gasUsed:          0,
		currentDepth:     0,
		hisResult:        make([]*callContractResult, 0),
		blockVersion:     blockVersion,
	}
}

// Storage interface for smart contracts
type txSimContextImpl struct {
	txExecSeq        int
	txResult         *common.Result
	txRWSet          *common.TxRWSet
	tx               *common.Transaction
	txReadKeyMap     map[string]*common.TxRead
	txWriteKeyMap    map[string]*common.TxWrite
	txWriteKeySql    []*common.TxWrite
	txWriteKeyDdlSql []*common.TxWrite // record ddl vm run success or failure
	snapshot         protocol.Snapshot
	vmManager        protocol.VmManager
	gasUsed          uint64 // only for callContract
	currentDepth     int
	currentResult    []byte
	hisResult        []*callContractResult
	sqlRowCache      map[int32]protocol.SqlRows
	kvRowCache       map[int32]protocol.StateIterator
	blockVersion     uint32
	keyIndex         int
}

type callContractResult struct {
	contractName string
	method       string
	param        map[string][]byte
	depth        int
	gasUsed      uint64
	result       []byte
}

func (s *txSimContextImpl) Get(contractName string, key []byte) ([]byte, error) {
	// Get from write set
	if value, done := s.getFromWriteSet(contractName, key); done {
		s.putIntoReadSet(contractName, key, value)
		return value, nil
	}

	// Get from read set
	if value, done := s.getFromReadSet(contractName, key); done {
		return value, nil
	}

	// Get from db
	var value []byte
	var err error
	if value, err = s.snapshot.GetKey(s.txExecSeq, contractName, key); err != nil {
		return nil, err
	}
	// if get from db success, put into read set
	s.putIntoReadSet(contractName, key, value)
	return value, nil
}
func (s *txSimContextImpl) Put(contractName string, key []byte, value []byte) error {
	s.putIntoWriteSet(contractName, key, value)
	return nil
}

func (s *txSimContextImpl) PutRecord(contractName string, value []byte, sqlType protocol.SqlType) {
	txWrite := &common.TxWrite{
		Key:          []byte(s.getSqlKey()),
		Value:        value,
		ContractName: contractName,
	}
	s.txWriteKeySql = append(s.txWriteKeySql, txWrite)
	if sqlType == protocol.SqlTypeDdl {
		s.txWriteKeyDdlSql = append(s.txWriteKeyDdlSql, txWrite)
	}
}

func (s *txSimContextImpl) getSqlKey() string {
	s.keyIndex++
	return "#sql#" + s.tx.Payload.TxId + "#" + strconv.Itoa(s.keyIndex)
}

func (s *txSimContextImpl) Del(contractName string, key []byte) error {
	s.putIntoWriteSet(contractName, key, nil)
	return nil
}

func (s *txSimContextImpl) Select(contractName string, startKey []byte, limit []byte) (protocol.StateIterator, error) {
	// 将来需要把txRwSet的最新状态填充到Iter中去，覆盖或者替换，才是完整的最新的Iter，否则就只是数据库的状态
	return s.snapshot.GetBlockchainStore().SelectObject(contractName, startKey, limit)
}

func (s *txSimContextImpl) GetCreator(contractName string) *acPb.Member {
	contract, err := s.GetContractByName(contractName)
	if err != nil {
		//TODO log
		return nil
	}
	return contract.Creator
}

func (s *txSimContextImpl) GetSender() *acPb.Member {
	return s.tx.Sender.GetSigner()
}

func (s *txSimContextImpl) putIntoReadSet(contractName string, key []byte, value []byte) {
	s.txReadKeyMap[constructKey(contractName, key)] = &common.TxRead{
		Key:          key,
		Value:        value,
		ContractName: contractName,
		Version:      nil,
	}
}

func (s *txSimContextImpl) putIntoWriteSet(contractName string, key []byte, value []byte) {
	s.txWriteKeyMap[constructKey(contractName, key)] = &common.TxWrite{
		Key:          key,
		Value:        value,
		ContractName: contractName,
	}
}

func (s *txSimContextImpl) getFromReadSet(contractName string, key []byte) ([]byte, bool) {
	if txRead, ok := s.txReadKeyMap[constructKey(contractName, key)]; ok {
		return txRead.Value, true
	}
	return nil, false
}

func (s *txSimContextImpl) getFromWriteSet(contractName string, key []byte) ([]byte, bool) {
	if txWrite, ok := s.txWriteKeyMap[constructKey(contractName, key)]; ok {
		return txWrite.Value, true
	}
	return nil, false
}

// Get the corresponding transaction
func (s *txSimContextImpl) GetTx() *common.Transaction {
	return s.tx
}

// Get blockchain storage
func (s *txSimContextImpl) GetBlockchainStore() protocol.BlockchainStore {
	return s.snapshot.GetBlockchainStore()
}

// GetAccessControl get access control service
func (s *txSimContextImpl) GetAccessControl() (protocol.AccessControlProvider, error) {
	if s.vmManager.GetAccessControl() == nil {
		return nil, errors.New("access control for tx sim context is nil")
	}
	return s.vmManager.GetAccessControl(), nil
}

// Get organization service
func (s *txSimContextImpl) GetChainNodesInfoProvider() (protocol.ChainNodesInfoProvider, error) {
	if s.vmManager.GetChainNodesInfoProvider() == nil {
		return nil, errors.New("chainNodesInfoProvider for tx sim context is nil, may be running in singleton mode")
	}
	return s.vmManager.GetChainNodesInfoProvider(), nil
}

// GetTxRWSet return current transaction read write set
func (s *txSimContextImpl) GetTxRWSet(runVmSuccess bool) *common.TxRWSet {
	if s.txRWSet != nil {
		return s.txRWSet
	}
	s.txRWSet = &common.TxRWSet{
		TxId:     s.tx.Payload.TxId,
		TxReads:  nil,
		TxWrites: nil,
	}

	// read set
	{
		txIds := make([]string, 0, len(s.txReadKeyMap))
		for txId := range s.txReadKeyMap {
			txIds = append(txIds, txId)
		}
		sort.Strings(txIds)
		for _, k := range txIds {
			s.txRWSet.TxReads = append(s.txRWSet.TxReads, s.txReadKeyMap[k])
		}
	}

	// write set
	if runVmSuccess {
		txIds := make([]string, 0, len(s.txWriteKeyMap))
		for txId := range s.txWriteKeyMap {
			txIds = append(txIds, txId)
		}
		sort.Strings(txIds)
		for _, k := range txIds {
			s.txRWSet.TxWrites = append(s.txRWSet.TxWrites, s.txWriteKeyMap[k])
		}
		// sql nil key tx writes
		s.txRWSet.TxWrites = append(s.txRWSet.TxWrites, s.txWriteKeySql...)
	} else {
		// ddl sql tx writes
		s.txRWSet.TxWrites = s.txWriteKeyDdlSql
	}
	return s.txRWSet
}

// Get the height of the corresponding block
func (s *txSimContextImpl) GetBlockHeight() uint64 {
	return s.snapshot.GetBlockHeight()
}

func (s *txSimContextImpl) GetBlockProposer() *acPb.Member {
	return s.snapshot.GetBlockProposer()
}

// Obtain the corresponding transaction execution sequence
func (s *txSimContextImpl) GetTxExecSeq() int {
	return s.txExecSeq
}

// set the corresponding transaction execution sequence
func (s *txSimContextImpl) SetTxExecSeq(txExecSeq int) {
	s.txExecSeq = txExecSeq
}

// Get the tx result
func (s *txSimContextImpl) GetTxResult() *common.Result {
	return s.txResult
}

// Set the tx result
func (s *txSimContextImpl) SetTxResult(txResult *common.Result) {
	s.txResult = txResult
}

// Cross contract call
func (s *txSimContextImpl) CallContract(contract *common.Contract, method string, byteCode []byte,
	parameter map[string][]byte, gasUsed uint64, refTxType common.TxType) (
	*common.ContractResult, common.TxStatusCode) {
	s.gasUsed = gasUsed
	s.currentDepth = s.currentDepth + 1
	if s.currentDepth > protocol.CallContractDepth {
		contractResult := &common.ContractResult{
			Code:    uint32(1),
			Result:  nil,
			Message: fmt.Sprintf("CallContract too depth %d", s.currentDepth),
		}
		return contractResult, common.TxStatusCode_CONTRACT_TOO_DEEP_FAILED
	}
	if s.gasUsed > protocol.GasLimit {
		contractResult := &common.ContractResult{
			Code:    uint32(1),
			Result:  nil,
			Message: fmt.Sprintf("There is not enough gas, gasUsed %d GasLimit %d ", gasUsed, int64(protocol.GasLimit)),
		}
		return contractResult, common.TxStatusCode_CONTRACT_FAIL
	}
	if len(byteCode) == 0 {
		dbByteCode, err := s.GetContractBytecode(contract.Name)
		if err != nil {
			return nil, common.TxStatusCode_CONTRACT_FAIL
		}
		byteCode = dbByteCode
	}
	r, code := s.vmManager.RunContract(contract, method, byteCode, parameter, s, s.gasUsed, refTxType)

	result := callContractResult{
		depth:        s.currentDepth,
		gasUsed:      s.gasUsed,
		result:       r.Result,
		contractName: contract.Name,
		method:       method,
		param:        parameter,
	}
	s.hisResult = append(s.hisResult, &result)
	s.currentResult = r.Result
	s.currentDepth = s.currentDepth - 1
	return r, code
}

// Obtain the execution result of current contract (cross contract)
func (s *txSimContextImpl) GetCurrentResult() []byte {
	return s.currentResult
}

// Get contract call depth
func (s *txSimContextImpl) GetDepth() int {
	return s.currentDepth
}

func constructKey(contractName string, key []byte) string {
	return contractName + constructKeySeparator + string(key)
}

func (s *txSimContextImpl) SetStateSqlHandle(index int32, rows protocol.SqlRows) {
	// 当前交易总是串行执行，故不需要加锁
	s.sqlRowCache[index] = rows
}

func (s *txSimContextImpl) GetStateSqlHandle(index int32) (protocol.SqlRows, bool) {
	data, ok := s.sqlRowCache[index]
	return data, ok
}

func (s *txSimContextImpl) SetStateKvHandle(index int32, rows protocol.StateIterator) {
	// 当前交易总是串行执行，故不需要加锁
	s.kvRowCache[index] = rows
}

func (s *txSimContextImpl) GetStateKvHandle(index int32) (protocol.StateIterator, bool) {
	data, ok := s.kvRowCache[index]
	return data, ok
}
func (s *txSimContextImpl) GetBlockVersion() uint32 {
	return s.blockVersion
}

func (s *txSimContextImpl) GetContractByName(name string) (*common.Contract, error) {
	return s.snapshot.GetBlockchainStore().GetContractByName(name)
}

//GetContractBytecode get contract bytecode
func (s *txSimContextImpl) GetContractBytecode(name string) ([]byte, error) {
	return s.snapshot.GetBlockchainStore().GetContractBytecode(name)
}
